/*************************************************************************
 
ftpsfv.c
 
Copyright (c) 2000-2003, webbie@ipfw.org
      2003 tl@bunker-werk.net
All rights reserved.
 
Derived from bsd-sfv.
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
 
Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.
 
Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 
*************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <dirent.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#include "version.h"
#include "config.h"

#define FNAMELEN		250
#define MAXSFVFILE		1024
#define FTPSFV_VERSION		"ftpSFV for UNiX, " VERSION
#define MISSINGTAG 		".missing"
#define BADTAGBEGIN		"-[BAD]-["
#define BADTAGEND		"]-"

#define UNKNOWN			0
#define OK			1
#define BAD			2
#define MISSING			3
#define CHECKING		4

char SITENAME[FNAMELEN] = "ftpSFV";
char TMPFILENAME[FNAMELEN] = ".tmp";
char cCOMPLETETAG[FNAMELEN] = "[100\%]-[-all.files.CRC.ok-]";
char cWORKINGTAG[FNAMELEN] = "[000\%]-[-working-]";
char OLDTAG[FNAMELEN];
char NEWTAG[FNAMELEN];
mode_t FILEPERM = 0777;
mode_t DIRPERM = 0777;
uid_t USER = -1;
gid_t GROUP = -1;

typedef struct sfvtable
  {
    char filename[FNAMELEN];
    int crc;
    int found;
    int status;
    time_t lastmodification;
  }
SFVTABLE;

SFVTABLE sfvTable[MAXSFVFILE];
SFVTABLE tmpsfvTable[MAXSFVFILE];
int numfiles = 0;
int current_file = 0;


long UpdateCRC(unsigned long CRC, const char *buffer, long count)
{
  /*
   * Note: if you want to know how CRC32-checking works, I
   * recommend grabbing any old (mid 90's) Z-Modem source.
   * There is not much you can change in this function, so
   * if you need a CRC32-check yourself, feel free to rip.
   */
  unsigned long CRCTABLE[] = {
                               0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419,
                               0x706af48f,
                               0xe963a535, 0x9e6495a3, 0x0edb8832, 0x79dcb8a4, 0xe0d5e91e,
                               0x97d2d988,
                               0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91, 0x1db71064,
                               0x6ab020f2,
                               0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551,
                               0x83d385c7,
                               0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f,
                               0x63066cd9,
                               0xfa0f3d63, 0x8d080df5, 0x3b6e20c8, 0x4c69105e, 0xd56041e4,
                               0xa2677172,
                               0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b, 0x35b5a8fa,
                               0x42b2986c,
                               0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf,
                               0xabd13d59,
                               0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5,
                               0x56b3c423,
                               0xcfba9599, 0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2,
                               0xb10be924,
                               0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 0x76dc4190,
                               0x01db7106,
                               0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5,
                               0xe8b8d433,
                               0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb,
                               0x086d3d2d,
                               0x91646c97, 0xe6635c01, 0x6b6b51f4, 0x1c6c6162, 0x856530d8,
                               0xf262004e,
                               0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6,
                               0x12b7e950, 

                               0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3,
                               0xfbd44c65,
                               0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541,
                               0x3dd895d7,
                               0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a, 0x346ed9fc, 0xad678846,
                               0xda60b8d0,
                               0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c,
                               0x270241aa,
                               0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409,
                               0xce61e49f,
                               0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17,
                               0x2eb40d81,
                               0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6, 0x03b6e20c,
                               0x74b1d29a,
                               0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683, 0xe3630b12,
                               0x94643b84,
                               0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27,
                               0x7d079eb1,
                               0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d,
                               0x806567cb,
                               0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a,
                               0x67dd4acc,
                               0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5, 0xd6d6a3e8,
                               0xa1d1937e,
                               0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd,
                               0x48b2364b,
                               0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3,
                               0xa867df55,
                               0x316e8eef, 0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0,
                               0x5268e236,
                               0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe,
                               0xb2bd0b28,
                               0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b,
                               0x5bdeae1d,
                               0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9,
                               0xeb0e363f,
                               0x72076785, 0x05005713, 0x95bf4a82, 0xe2b87a14, 0x7bb12bae,
                               0x0cb61b38,
                               0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4,
                               0xf1d4e242,
                               0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1,
                               0x18b74777,
                               0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff,
                               0xf862ae69,
                               0x616bffd3, 0x166ccf45, 0xa00ae278, 0xd70dd2ee, 0x4e048354,
                               0x3903b3c2,
                               0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db, 0xaed16a4a,
                               0xd9d65adc,
                               0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f,
                               0x30b5ffe9,
                               0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605,
                               0xcdd70693,
                               0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8, 0x5d681b02,
                               0x2a6f2b94,
                               0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
                             };

  if (buffer && count)
    {
      do
        {
          CRC =
            ((CRC >> 8) & 0xFFFFFF) ^
            CRCTABLE[(unsigned char) ((CRC & 0xff) ^ *buffer++)];
        }
      while (--count);

    }
  return CRC;
}

/* ^ UpdateCRC ^ */



long GetFileCRC(char *filename)
{
  unsigned long crc = 0xffffffff;
  FILE *f;
  long totalread = 0;
  long localread;

  /*
   * Note: different buffer sizes may result in noticable 
   * different performance depending on system, so feel
   * free to modify.
   */
#define BUFFERSIZE 65536*16

  char buffer[BUFFERSIZE];

  if ((f = fopen(filename, "rb")) != NULL)
    {
      do
        {
          if ((localread = fread(buffer, 1, BUFFERSIZE, f)))
            {
              crc = UpdateCRC(crc, buffer, localread);
              totalread = totalread + localread;
            }
        }
      while (localread > 0);
      fclose(f);

      crc = crc ^ 0xffffffff;
    }
  else
    {			/* error opening file */

      fprintf(stderr, "\nFatal error: cannot read file: %s\n", filename);
    }

  return crc;
}


/* lookup UID from user */
uid_t lookupUID(char *user)
{
  struct passwd *u;

  errno = 0;
  u = getpwnam(user);
  if (!u)
    return -1;
  return (u->pw_uid);
}


/* lookup GID from group */
gid_t lookupGID(char *group)
{
  struct group *g;

  errno = 0;
  g = getgrnam(group);
  if (!g)
    return -1;
  return (g->gr_gid);

}


char *basename(const char *path)
{				/* returns the filename portion of the string */
  static char bname[MAXPATHLEN];
  register const char *endp, *startp;

  if (path == NULL || *path == '\0')
    {
      (void) strncpy(bname, ".", sizeof bname);
      return (bname);
    }
  endp = path + strlen(path) - 1;
  while (endp > path && *endp == '/')
    endp--;
  if (endp == path && *endp == '/')
    {
      (void) strncpy(bname, "/", sizeof bname);
      return (bname);
    }
  startp = endp;
  while (startp > path && *(startp - 1) != '/')
    startp--;
  if (endp - startp + 2 > sizeof(bname))
    {
      errno = ENAMETOOLONG;
      return (NULL);
    }
  strncpy(bname, startp, endp - startp + 2);
  return (bname);
}

int FileExists(char *name)
{
  FILE *testfile = NULL;

  testfile = fopen(name, "rb");
  if (testfile != NULL)
    {
      fclose(testfile);
      return 1;
    }
  return 0;
}


void FileCreate(char *filename)
{
  FILE *crapfile;

  crapfile = fopen(filename, "w+");
  if (crapfile != NULL)
    {
      fclose(crapfile);
      chmod(filename, FILEPERM);
      chown(filename, USER, GROUP);	
    }
}

void DirCreate(char *dirname)
{
  if (mkdir(dirname, S_IRWXU | S_IRWXG | S_IRWXO) == 0)
    chmod(dirname, DIRPERM);
    chown(dirname, USER, GROUP);
}

int CheckFileExists(char *name, SFVTABLE sfvTable[], int n)
{
  while (n > 0)
    {
      if (!strncasecmp(sfvTable[--n].filename, name, FNAMELEN))
        return n;
    }
  return (-1);
}

void usage(char *prog)
{
  printf("\nUsage: %s mode [options] sfv-filename \n\n", prog);
  printf("Options:\n");
  printf("-s site-name\tUse site-name as site-name, default ftpSFV\n");
  printf("-t file-name\tUse file-name as name for tmp-file, default .tmp\n");
  printf("-u user-name\tUse the user user-name as owner of created files/dirs\n");
  printf("\t\tOnly works if run as root\n");
  printf("\t\tdefault is the user which runs ftpsfv\n");
  printf("-g group-name\tUse the group groupname as owner of created files/dirs\n");
  printf("\t\tOnly works if run as root\n");
  printf("\t\tdefault is the group which runs ftpsfv\n");
  printf("-f permission\tSets the UMASK for files created, default 0777\n");
  printf("-d permission\tSets the UMASK for directories created, default 0777\n");
  printf("\n");
  return;
}

int read_sfv(char *sfvname)
{
  FILE *sfvfile;
  char sfvline[FNAMELEN];	// current line in SFV file
  char *token;
  int tmp;
  char chtmp[FNAMELEN];
  size_t stmp = 0;
  int numfiles;

  sfvfile = fopen(sfvname, "r");
  if (sfvfile == NULL)
    {
      fprintf(stderr, "Failed to open %s\n", sfvname);
      exit(1);
    }
  numfiles = 0;
  while (fgets(sfvline, FNAMELEN, sfvfile) != NULL)
    {

      // Skip comments, blank lines, other crap
      if (sfvline[0] != ';' && sfvline[0] != '#'
          && sfvline[0] != ' ' && strlen(sfvline) > 8)
        {
          token = strtok(sfvline, "\n\r");
          if ((tmp = strlen(token)) > 8)
            {
              strncpy(chtmp, &token[tmp - 8], 8);
              sscanf(chtmp, "%X", &sfvTable[numfiles].crc);
              strncpy(sfvTable[numfiles].filename, &token[0], tmp - 8);
              for (stmp = strlen(sfvTable[numfiles].filename) - 1;
                   stmp >= 0; stmp--)
                {
                  if (sfvTable[numfiles].filename[stmp] == ' ')
                    sfvTable[numfiles].filename[stmp] = '\0';
                  else
                    break;
                }

              sfvTable[numfiles].status = UNKNOWN;
              numfiles++;
              if (numfiles >= MAXSFVFILE)
                {
                  printf("ftpsfv cannot handle more than %d files\n",
                         MAXSFVFILE);
                  exit(1);
                }
            }
        }
    };
  fclose(sfvfile);
  return numfiles;
}

void get_old_dir_tag(void)
{
  DIR *dirp;
  struct dirent *dp;
  struct stat statbuf;
  int cnt;

  OLDTAG[0] = '\0';
  dirp = opendir(".");
  while ((dp = readdir(dirp)) != NULL)
    {
      if ((dp->d_type == DT_DIR || dp->d_type == DT_UNKNOWN)
          && (dp->d_name[0] == '[')
          && (dp->d_name[strlen(SITENAME) + 1] == ']')
          && (dp->d_name[strlen(SITENAME) + 2] == '-')
          && !strncmp(dp->d_name + 1, SITENAME, strlen(SITENAME)))
        {
          strcpy(OLDTAG, dp->d_name);
        }

      if ((dp->d_type == DT_DIR || dp->d_type == DT_UNKNOWN)
          &&
          ((strlen(dp->d_name) ==
            strlen(cCOMPLETETAG) + strlen(SITENAME) + 3)
           && (dp->d_name[0] == '[')
           && (dp->d_name[strlen(SITENAME) + 1] == ']')
           && (dp->d_name[strlen(SITENAME) + 2] == '-')
           && !strncmp(dp->d_name + 1, SITENAME, strlen(SITENAME)))
          && !strncmp(dp->d_name + strlen(SITENAME) + 3, cCOMPLETETAG,
                      strlen(cCOMPLETETAG)))
        {
          printf("Nothing to do. All files already checked.\n");
          exit(0);
        }

      if (dp->d_type == DT_REG || dp->d_type == DT_UNKNOWN)
        {
          cnt = CheckFileExists(dp->d_name, sfvTable, numfiles);
          if (cnt >= 0)
            {
              sfvTable[cnt].found = 1;
              strcpy(sfvTable[cnt].filename, dp->d_name);
              if (stat(sfvTable[cnt].filename, &statbuf))
                sfvTable[cnt].lastmodification = statbuf.st_mtime;
              else
                sfvTable[cnt].lastmodification = 0;
            }
        }
    }
  (void) closedir(dirp);
}

int read_tmpfile(SFVTABLE tmpsfvTable[], int updatefname)
{
  FILE *tmpfile;
  char *token;
  char tmpline[FNAMELEN];	// current line in tmp file
  int tmp;
  int cnt = 0;

  /* read tmp-file to see which files where already checked */
  tmpfile = fopen(TMPFILENAME, "r");
  if (tmpfile != NULL)
    {
      while (fgets(tmpline, FNAMELEN, tmpfile) != NULL)
        {
          cnt++;
          token = strtok(tmpline, "\t");
          if ((tmp = CheckFileExists(token, sfvTable, numfiles)) != -1)
            {
              if (updatefname)
                strcpy(tmpsfvTable[tmp].filename, token);
              token = strtok(NULL, " \n\r\t");
              if (strcasecmp(token, "OK") == 0)
                tmpsfvTable[tmp].status = OK;
              if (strcasecmp(token, "BAD") == 0)
                tmpsfvTable[tmp].status = BAD;
              if (strcasecmp(token, "MISSING") == 0)
                tmpsfvTable[tmp].status = MISSING;
              if (strcasecmp(token, "UNKNOWN") == 0)
                tmpsfvTable[tmp].status = UNKNOWN;
              if (strcasecmp(token, "CHECKING") == 0)
                tmpsfvTable[tmp].status = CHECKING;
            }
        }
      fclose(tmpfile);
    }
  else
    return -1;
  return cnt;
}

void init_tmpfile(void)
{
  FILE *tmpfile;
  int cnt;

  tmpfile = fopen(TMPFILENAME, "w");
  if (tmpfile == NULL)
    {
      fprintf(stderr, "Failed to open tmp-file %s\n", TMPFILENAME);
    }
  else
    {
      for (cnt = 0; cnt < numfiles; cnt++)
        {
          fputs(sfvTable[cnt].filename, tmpfile);
          fputs("\tUNKNOWN\n", tmpfile);
        }
      fclose(tmpfile);
      chmod(TMPFILENAME, FILEPERM);
      chown(TMPFILENAME, USER, GROUP);	
    }
}

void update_tmpfile(int where, int value)
{
  FILE *tmpfile;
  SFVTABLE tmpsfvTable[MAXSFVFILE];
  int cnt;

  memset(tmpsfvTable, 0, sizeof(tmpsfvTable));
  memcpy(tmpsfvTable, sfvTable, sizeof(sfvTable));
  read_tmpfile(tmpsfvTable, 0);
  
  tmpfile = fopen(TMPFILENAME, "w");
  if (tmpfile == NULL)
    {
      fprintf(stderr, "Failed to open tmp-file %s\n", TMPFILENAME);
    }
  else
    {
      for (cnt = 0; cnt < numfiles; cnt++)
        {
          fputs(tmpsfvTable[cnt].filename, tmpfile);
          if (cnt == where)
            tmpsfvTable[cnt].status = value;
          switch (tmpsfvTable[cnt].status)
            {
            case OK:
              fputs("\tOK\n", tmpfile);
              break;
            case MISSING:
              fputs("\tMISSING\n", tmpfile);
              break;
            case BAD:
              fputs("\tBAD\n", tmpfile);
              break;
            case CHECKING:
              fputs("\tCHECKING\n", tmpfile);
              break;
            default:
              fputs("\tUNKNOWN\n", tmpfile);
              break;
            }
        }

    }
  fclose(tmpfile);
  chmod(TMPFILENAME, FILEPERM);
  chown(TMPFILENAME, USER, GROUP);	
}

char *status_str(int status)
{
  switch (status)
    {
    case OK:
      return("OK");
      break;
    case MISSING:
      return("MISSING");
      break;
    case BAD:
      return("BAD");
      break;
    case CHECKING:
      return("CHECKING");
    default:
      return("UNKNOWN");
      break;
    }
  return("none");
}

int main(int argc, char *argv[])
{
  char sfvname[FNAMELEN];	// name of .sfv file to use
  SFVTABLE tmpsfvTable[MAXSFVFILE];

  long mycrc = 0;
  int cnt;
  int num;
  char crap[FNAMELEN];
  char precent[5];
  char tmp[FNAMELEN];
  int badfiles = 0;
  int missingfiles = 0;
  int unknownfiles = 0;
  char ch;
  int percent;
  char *endp;
  struct passwd *pw;
  struct group *grp;
	
  // output program name & version number
  printf("%s\n", FTPSFV_VERSION);

  // parse command line
  cnt = 0;
  while ((ch = getopt(argc, argv, "v?hs:t:u:f:g:d:")) != -1)
    {
      switch (ch)
        {
        case 'v':		// print version and exit
          return 0;
          break;

        case 's':		// set site-name
          snprintf(SITENAME, FNAMELEN, "%s", optarg);
          printf("Using site-name <%s>\n", SITENAME);
          break;
        
	case 't':		// set tmp-file name
          snprintf(TMPFILENAME, FNAMELEN, "%s", optarg);
          printf("Using tmp-file <%s>\n", TMPFILENAME);
          break;

	case 'u':		// set the user created files/dirs belong to
	  snprintf(tmp, FNAMELEN, "%s", optarg);
	  if ((USER = lookupUID(tmp)) == -1)
	    {
	      USER = getuid();
	      pw = getpwuid(getuid());
	      sprintf(tmp, "%s", pw->pw_name);
	    }
	  printf("Using user <%s> UID <%i>\n", tmp, USER);
	  break;

	case 'g':		// set the group created files/dirs belong to
	  snprintf(tmp, FNAMELEN, "%s", optarg);
	  if ((GROUP = lookupGID(tmp)) == -1)
	    {
	      GROUP = getgid();
	      grp = getgrgid(getgid());
	      sprintf(tmp, "%s", grp->gr_name);
	    }
	  printf("Using group <%s> GID <%i>\n", tmp, GROUP);
	  break;
	    
	case 'f':		// set the permissions to the files created
	  FILEPERM = (mode_t) strtol(optarg, &endp, 8);
	  if (endp && *endp)
	    FILEPERM = 0777;
	  printf("Using file permission mask <%04o>\n", FILEPERM);
	  break;
	
	case 'd':		// set the permissions to the dirs created
	  DIRPERM = (mode_t) strtol(optarg, &endp, 8);
	  if (endp && *endp)
	    DIRPERM = 0777;
	  printf("Using directory permission mask <%04o>\n", DIRPERM);
	  break;
       
	case '?':
        case 'h':
          usage(argv[0]);
          return 1;
          break;
        }
    }

  if (argc == optind)
    {
      usage(argv[0]);
      return 1;
    }
  argc -= optind;
  argv += optind;

  strncpy(sfvname, argv[0], FNAMELEN);
  argc--;
  argv++;

  printf("\nProcessing complete check of %s ...\n", sfvname);

  strcat(TMPFILENAME, ".");
  strncpy(tmp, basename(sfvname), strlen(basename(sfvname))-4);
  strncat(TMPFILENAME, tmp, FNAMELEN - strlen(TMPFILENAME));

  numfiles = read_sfv(sfvname);

  get_old_dir_tag();

  if (read_tmpfile(sfvTable, 0) == -1)
    init_tmpfile();
  else
    printf("Found .tmp file.\n");

  for (current_file = 0; current_file < numfiles; current_file++)
    {
      read_tmpfile(sfvTable, 0);
      percent = 0;
      if (numfiles > 0)
        {
          percent = ((numfiles - current_file) * 100) / numfiles;
          percent = 100 - percent;
        }

      sprintf(precent, "%.3d", percent);
      strcpy(NEWTAG, cWORKINGTAG);
      NEWTAG[1] = precent[0];
      NEWTAG[2] = precent[1];
      NEWTAG[3] = precent[2];

      snprintf(tmp, FNAMELEN, "[%s]-%s", SITENAME, NEWTAG);
      strcpy(NEWTAG, tmp);

      if (strcmp(OLDTAG, NEWTAG))
        {
          rmdir(OLDTAG);
	  DirCreate(NEWTAG);
          strcpy(OLDTAG, NEWTAG);
        }

      printf("\nTesting %s ... status = %s listed = 0x%08X ... ",
             sfvTable[current_file].filename,
             status_str(sfvTable[current_file].status),
             sfvTable[current_file].crc);

      fflush(stdout);

      update_tmpfile(current_file, CHECKING);

      if ((sfvTable[current_file].status != OK) &&
          (sfvTable[current_file].status != CHECKING))
        mycrc = GetFileCRC(sfvTable[current_file].filename);
      else
        mycrc = sfvTable[current_file].crc;

      printf("local = 0x%08lX ... ", mycrc);
      if (mycrc == sfvTable[current_file].crc)
        {
          sfvTable[current_file].status = OK;
          printf("OK\n");
          snprintf(crap, FNAMELEN,
                   "%s%s", sfvTable[current_file].filename,
                   MISSINGTAG);
          unlink(crap);
          snprintf(crap, FNAMELEN,
                   "%s%s%s", BADTAGBEGIN,
                   sfvTable[current_file].filename, BADTAGEND);
          unlink(crap);
        }
      else
        {
          if (!sfvTable[current_file].found)
            {
              missingfiles++;
              sfvTable[current_file].status = MISSING;
              printf("MISSING\n");
              snprintf(crap, FNAMELEN,
                       "%s%s%s",
                       BADTAGBEGIN, sfvTable[current_file].filename,
                       BADTAGEND);
              unlink(crap);
              snprintf(crap, FNAMELEN,
                       "%s%s", sfvTable[current_file].filename,
                       MISSINGTAG);
              FileCreate(crap);
            }
          else
            {
              badfiles++;
              sfvTable[current_file].status = BAD;
              printf("BAD\n");
              snprintf(crap, FNAMELEN,
                       "%s%s", sfvTable[current_file].filename,
                       MISSINGTAG);
              unlink(crap);
              snprintf(crap, FNAMELEN,
                       "%s%s%s",
                       BADTAGBEGIN, sfvTable[current_file].filename,
                       BADTAGEND);
              FileCreate(crap);
            }
        }
      update_tmpfile(current_file, sfvTable[current_file].status);
      fflush(stdout);
    }

  strcpy(NEWTAG, cCOMPLETETAG);

  missingfiles = 0;
  badfiles = 0;
  unknownfiles = 0;

  num = read_tmpfile(tmpsfvTable, 0);
  if  (num > 0)
    for (cnt = 0; cnt < num; cnt++)
      {
        switch (tmpsfvTable[cnt].status)
          {
          case MISSING:
            missingfiles++;
            break;
          case BAD:
            badfiles++;
            break;
          case UNKNOWN:
            unknownfiles++;
            break;
          }
      }

  if (!missingfiles && !badfiles && !unknownfiles)
    unlink(TMPFILENAME);

  if (missingfiles || badfiles)
    snprintf(NEWTAG, FNAMELEN,
             "[%i.files.OK]-[%i.files.BAD]-[%i.files.MISSING]",
             numfiles - (missingfiles + badfiles), badfiles,
             missingfiles);

  snprintf(tmp, FNAMELEN, "[%s]-%s", SITENAME, NEWTAG);
  strcpy(NEWTAG, tmp);

  if (strcmp(OLDTAG, NEWTAG))
    {
      rmdir(OLDTAG);
      DirCreate(NEWTAG);
    }

  printf("\n%d file(s) tested - %d OK - %d bad - %d missing ...\n\n",
         numfiles, numfiles - (badfiles + missingfiles), badfiles,
         missingfiles);

  if (missingfiles)
    return 2;

  if (badfiles)
    return 1;

  return 0;
}
